/***************************************************************************
 * LPRng - An Extended Print Spooler System
 *
 *
 * event handling for ipp subscriptions
 * copyright 2021 Vaclav Michalek, Jonit Laberatory of Optics, Olomouc 
 *
 * See LICENSE for conditions of use.
 *
 ***************************************************************************/

#include "lp.h"
#include "linelist.h"
#include "utilities.h"
#include "event.h"
#include "ipp_events.h"


/*remove event pipes for ended children*/
int Remove_closed_pipes( struct line_list *clnt_fd )
{

	/*should not happen, but who knows*/
	int i, fd, status = 0;

	/*DEBUG1( "Remove_closed_pipes: start" );*/
	if( DEBUGL3 ) Dump_line_list ("Clnt fds: ", clnt_fd);

	if (clnt_fd->count > Process_list.count ) {
		for ( (i = clnt_fd->count),--i; i >= 0; --i) {
			fd = strtol(clnt_fd->list[i], 0, 0 ); if (!fd) continue;
			status = Event_send2(fd, E_NONE, 0, 0);
			if (status != 0) {
				DEBUG2( "Remove_closed_pipes: fd %d not writable, removing ", fd );
				Remove_line_list(clnt_fd, i);
				close(fd);
				/*remove running/dead event notifications*/
				Ips_delete_wait_client(fd);
			}
		}
	}

	return clnt_fd->count;
}


int Event_send2( int fd, int evno, int p1, int p2 )
{

	char buf[LINEBUFFER];
	int bytes, err;

	DEBUG2( "Event_send2: fd %d start", fd );

	memset(buf, 0, sizeof(buf));
	EddIntData(buf, sizeof(buf), evno);
	EddIntData(buf, sizeof(buf), p1);
	EddIntData(buf, sizeof(buf), p2);
	EddLf(buf, sizeof(buf));
	
	bytes = write(fd, buf, safestrlen(buf));
	err = bytes <= 0 ? errno : 0;
	DEBUG2( "Event_send2: fd %d event 0x%x p1 %d p2 %d sent %d bytes, errno %d", fd, evno, p1, p2, bytes, err);

	return bytes > 0 ? 0 : err;
}

int Event_dispatch( int srv_fd, struct line_list *clnt_fd )
{

	/*read and process events*/
	/*inspired by lpd.c/Read_server_status*/

	int status, sz, rs, lines, i, n, e;
	char bsz[SMALLBUFFER], *buffer;
	fd_set readfds;
	struct timeval timeval;
	struct line_list l, t;


	buffer = 0;

	DEBUG1( "Event_dispatch: starting" );


	Init_line_list(&l);
	Init_line_list(&t);

	while(1){
		rs = 0;
		FD_ZERO( &readfds );
		FD_SET( srv_fd, &readfds );
		memset(&timeval, 0, sizeof(timeval));
		status = select( srv_fd+1, &readfds, NULL, NULL, &timeval );
		DEBUG2( "Event_dispatch: select status %d", status);
		if( status == 0 ){
			rs = 1;
			break;
		} else if( status < 0 ){
			close(srv_fd);
			break;
		}
		/*get sz*/
		status = ok_read(srv_fd, bsz, 2*sizeof(size_t));
		if (status >= 0) bsz[status] = '\0'; else bsz[0] = '\0';
		sz = strtoul(bsz, NULL, 16);
		if (sz <= 0) sz = SMALLBUFFER;
		DEBUG2( "Event_dispatch: sz read status %d, sz %lu", status, (unsigned long)sz);

		/*get data*/
		FD_ZERO( &readfds );
		FD_SET( srv_fd, &readfds );
		memset(&timeval, 0, sizeof(timeval));
		if ( (status = select( srv_fd+1, &readfds, NULL, NULL, &timeval )) <=0) {
			if (!status) {
				DEBUG2("Event_dispatch: server read response data timeout");
				rs = 1;
				break;
			} else {
				e = errno;
				DEBUG2("Event_dispatch: server read reponse data error %d", e);
				rs = 1;
				break;
			}
		}

		if (buffer) free(buffer);
		buffer = malloc_or_die(sz + 1, __FILE__, __LINE__);

		status = ok_read(srv_fd, buffer, sz);
		DEBUG2( "Event_dispatch: data read status %d", status );
		if( status <= 0 ){
			close(srv_fd);
			break;
		}
		buffer[status] = 0;
		/*DEBUG1( "Event dispatch: read status %d '%s'", status, buffer );*/
		/*split [more] commands*/
		Split(&l,buffer,"\n",0,0,0,' ',0,0);
		if(DEBUGL1)Dump_line_list("Event_dispatch - input", &l );
		for( lines = 0; lines < l.count; ++lines ){
			/*split one comand*/
			Split(&t,l.list[lines],Whitespace,0,0,0,0,0,0);
			if( (t.count < 2) ) continue;
			switch (strtol(t.list[0], 0, 0))
			{
				case E_CLOSE:
					if( (t.count >=2) && !Find_first_key(clnt_fd, t.list[1], "=", &n)  ) {
						/*child closes pipe*/
						i = strtol(t.list[1], 0, 0);
						close(i);
						Remove_line_list(clnt_fd, n);
						DEBUG2( "Event dispatch: closed fd %d", i );
						/*remove closed subscription clients - should not be needed
						Ips_delete_wait_client(i);
						*/
					}
					break;
				case E_SUBS_CREATE:
					Ips_remove_subscription(NULL,NULL,0);
					Ips_subscription_create(&t);
					break;
				case E_SUBS_ATTRS:
					Ips_remove_subscription(NULL,NULL,0);
					Ips_subscription_attributes(&t);
					break;
				case E_SUBS_RENEW:
					Ips_remove_subscription(NULL,NULL,0);
					Ips_subscription_renew(&t);
					break;
				case E_SUBS_CANCEL:
					Ips_subscription_cancel(&t);
					break;
				case E_SUBS_GNTF:
					Ips_remove_subscription(NULL,NULL,0);
					Ips_get_notifications(&t);
					break;
				case E_SUBS_CNTF:
					/*notify wait mode close by client*/
					if (t.count>=2) {
						i = strtol(t.list[1], 0, 0);
						Ips_delete_wait_client(i);
					}
					break;
				case E_PRS_CHANGE:
				case E_JOB_CREATE:
				case E_JOB_CHANGE:
					Ips_event(&t);
					break;
			}
			Free_line_list(&t);
		}
		Free_line_list(&l);
	}
	Free_line_list(&t);
	Free_line_list(&l);

	if (buffer) free(buffer);

#ifdef DMALLOC
        {
                char *s = sbrk(0);
                int n = s - malloc_area;
                DEBUG1("lpd: EVENT_DISPATCH - sbrk 0x%lx, cnt %d", (long)s, clnt_fd.count );
                if( n > 1024 ){
                        extern int dmalloc_outfile_fd;
                        char buffer[128];
                        plp_snprintf(buffer,sizeof(buffer), "lpd: EVENT_DISPATCH sbrk reports 0x%lx, or %d more memory\n", (long)s, n );
                        Write_fd_str(dmalloc_outfile_fd, buffer );
                        dmalloc_log_unfreed();
                        DEBUG1( "lpd: EVENT_DISPATCH sbrk reports 0x%lx, or %d more memory", (long)s, n );
                        malloc_area += n;
                }
        }
#endif

	return rs;
}

int EddLf(char *str, int ssize) {

	int i ;
	size_t z;
	const char hx[] = "0123456789abcdef";

	i = safestrlen(str);
	plp_snprintf(str + i, ssize - i, "\n");
	str[ssize - 2] = '\n';
	str[ssize - 1] = '\0';
	z = 2*sizeof(size_t);
	if ( ssize - z > 1) {
		for ( (i = z - 1), z = safestrlen(str) - z; i >= 0; i--, z >>= 4)
			str[i] = hx[z & 0xf]; 
	}
	return 0;
}

int EddIntData(char *str, int ssize, int datum) {

	int i, j;

	i = safestrlen(str);
	if (i == 0) {
		/*space for size number in hex*/
		for (j = 0; (j - 2*sizeof(size_t) > 0) && (j < ssize - 2 ); j++) {
			str[j] = '0';
			j++;
			str[j] = '0';
		}
		if (j < ssize) str[j++] = ' ';
		if (j < ssize) str[j] = '\0'; else str[ssize-1] = '\0';
		i = safestrlen(str);
	}
	plp_snprintf(str + i, ssize - i, "%d ", datum);
	str[ssize - 1] = '\0';
	return 0;
}

int EddHexData(char *str, int ssize, const char *data, int datalen) {

	int i, j;

	str[ssize - 1] = '\0';
	if ((data) && (datalen > 0)) {
		for (i = 0, j = safestrlen(str); i < datalen && j < ssize-2 ; ++i, ++j, ++j) {
			plp_snprintf(str + j, ssize - 1, "%02x", (unsigned char)data[i]);
		}
		i = safestrlen(str);
		if (i + 1 < ssize ) strcat(str, " ");
	} else {
		i = safestrlen(str);
		if (i < ssize - 2) strcat(str, "N ");
	}

	return 0;
}

/*convert from text-hex and malloc*/
int ExmDataHex(char **data, const char *src) {

	int i,len;
	char hd[3];

	if (!src) {
		*data = NULL;
		return 0;
	}

	len = safestrlen(src)/2;
	hd[2] = '\0';

	*data = malloc_or_die(len + 1, __FILE__, __LINE__);

	for (i = 0, hd[2] = 0; i < len ; ++i ) {
		hd[0] = src[2*i];
		hd[1] = src[2*i+1];
		(*data)[i] = strtol(hd, NULL, 16) & 0xff;
	}
	(*data)[len] = '\0';
	return len;
}
